/*
 * Copyright (c) 2017 Ted Unangst <tedu@tedunangst.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
package main

import (
	"bytes"
	"compress/gzip"
	"encoding/json"
	"fmt"
	"io"
	"io/ioutil"
	"log"
	"net/http"
	"net/url"
	"regexp"
	"sort"
	"strconv"
	"strings"

	"github.com/andybalholm/cascadia"
	"golang.org/x/net/html"
)

var permittedtags = []string{"title", "div", "span", "h1", "h2", "h3", "h4", "h5", "h6",
	"table", "thead", "tbody", "th", "tr", "td",
	"p", "br", "pre", "code", "blockquote", "strong", "em", "b", "i", "s", "sup",
	"ol", "ul", "li"}
var bannedtags = []string{"script", "style"}

func init() {
	sort.Strings(permittedtags)
	sort.Strings(bannedtags)
}

func contains(array []string, tag string) bool {
	idx := sort.SearchStrings(array, tag)
	return idx < len(array) && array[idx] == tag
}

func getattr(node *html.Node, attr string) string {
	for _, a := range node.Attr {
		if a.Key == attr {
			return a.Val
		}
	}
	return ""
}

func jsonfind(jmap map[string]interface{}, keys []string) string {
	var ii interface{}
	ii = jmap
	for _, key := range keys {
		idx, err := strconv.Atoi(key)
		if err == nil {
			m := ii.([]interface{})
			ii = m[idx]
		} else {
			m := ii.(map[string]interface{})
			ii = m[key]
		}
	}
	return ii.(string)
}

func extractblogshit(blogscheme, bloghost string, bodybytes []byte) *html.Node {
	re := regexp.MustCompile("'postId': '([0-9]+)'")
	m := re.FindSubmatch(bodybytes)
	if m == nil {
		return nil
	}
	targ := fmt.Sprintf("%s://%s/feeds/posts/default/%s?alt=json", blogscheme, bloghost, m[1])
	log.Printf("now i'm getting real content from %s", targ)
	jr, err := http.Get(targ)
	if err != nil {
		log.Println("json post err", err)
		return nil
	}
	defer jr.Body.Close()
	decoder := json.NewDecoder(jr.Body)
	var jmap map[string]interface{}
	err = decoder.Decode(&jmap)
	if err != nil {
		log.Println("json decode error")
		return nil
	}
	body := jsonfind(jmap, []string{"entry", "content", "$t"})
	root, _ := html.Parse(strings.NewReader(body))
	return root
}

func isblogshit(bodybytes []byte) bool {
	re := regexp.MustCompile("'postId': '([0-9]+)'")
	m := re.FindSubmatch(bodybytes)
	if m == nil {
		return false
	}
	return true
}

func procscriptnode(w io.Writer, script *html.Node) {
	src := getattr(script, "src")
	if src != "" {
		scripturl, err := url.Parse(src)
		if err != nil {
			log.Println("frame err", err)
			return
		}
		if scripturl.Host == "gist.github.com" {
			path := scripturl.Path
			if strings.HasSuffix(path, ".js") {
				path = path[:len(path)-3] + "/raw"
			}
			scripturl.Path = path
			txt, err := http.Get(scripturl.String())
			if err != nil {
				log.Println("frame err", err)
				return
			}
			defer txt.Body.Close()
			var buf bytes.Buffer
			io.Copy(&buf, txt.Body)
			io.WriteString(w, fmt.Sprintf(
				`<pre style="border: 2px solid black; padding: 1em;">%s</pre>`,
				html.EscapeString(buf.String())))
		}
	}

}

func insertframecontents(w io.Writer, baseurl *url.URL, src string) {
	u, err := url.Parse(src)
	if err != nil {
		log.Println("frame err", err)
		return
	}
	targ := baseurl.ResolveReference(u)
	fr, err := http.Get(targ.String())
	if err != nil {
		log.Println("frame err", err)
		return
	}
	defer fr.Body.Close()
	if strings.HasPrefix(fr.Header.Get("Content-Type"), "text/html") {
		root, err := html.Parse(fr.Body)
		if err != nil {
			log.Println("frame err", err)
			return
		}
		sel := cascadia.MustCompile("script")
		scripts := sel.MatchAll(root)
		for _, script := range scripts {
			procscriptnode(w, script)
		}
		blocksel := cascadia.MustCompile("blockquote")
		quotes := blocksel.MatchAll(root)
		for _, quote := range quotes {
			clean(w, quote, nil)
		}
	}
}

func haschild(node *html.Node, css string) bool {
	sel := cascadia.MustCompile(css)
	return sel.MatchFirst(node) != nil
}

func clean(w io.Writer, node *html.Node, req *http.Request) {
	switch node.Type {
	case html.ElementNode:
		tag := node.Data
		switch {
		case tag == "a":
			io.WriteString(w, fmt.Sprintf(`<a href="%s">`,
				html.EscapeString(getattr(node, "href"))))
		case tag == "img":
			src := getattr(node, "src")
			if getattr(node, "data-native-src") != "" {
				src = getattr(node, "data-native-src")
				if strings.HasSuffix(src, "-1x-1.jpg") {
					src = src[:len(src)-9] + "1200x-1.jpg"
				}
			} else if getattr(node, "data-raw-src") != "" {
				src = getattr(node, "data-raw-src")
			}

			if strings.Contains(getattr(node, "class"), "progressiveMedia-thumbnail") {
				src = ""
			}
			if src != "" {
				io.WriteString(w, fmt.Sprintf(`<img src="%s">`,
					html.EscapeString(src)))
			}
		case tag == "picture":
			src := ""
			for c := node.FirstChild; c != nil; c = c.NextSibling {
				if c.Type == html.ElementNode && c.Data == "source" {
					attr := getattr(c, "data-srcset")
					if attr != "" {
						src = attr
					}
				}
			}
			if src != "" {
				io.WriteString(w, fmt.Sprintf(`<picture><img src="%s"></picture>`,
					html.EscapeString(src)))
				return
			}
		case tag == "figure":
			src := getattr(node, "itemid")
			if src != "" {
				if !haschild(node, "img") {
					io.WriteString(w, fmt.Sprintf(`<img src="%s">`,
						html.EscapeString(src)))
				}
			}
		case tag == "iframe":
			src := getattr(node, "src")
			if src != "" {
				insertframecontents(w, req.URL, src)
			}
		case tag == "script":
			procscriptnode(w, node)
			return
		case contains(permittedtags, tag):
			io.WriteString(w, fmt.Sprintf("<%s>", tag))
		case contains(bannedtags, tag):
			return
		}
	case html.TextNode:
		io.WriteString(w, html.EscapeString(node.Data))
	}
	for c := node.FirstChild; c != nil; c = c.NextSibling {
		clean(w, c, req)
	}
	if node.Type == html.ElementNode {
		tag := node.Data
		if tag == "a" || (contains(permittedtags, tag) && tag != "br") {
			io.WriteString(w, fmt.Sprintf("</%s>", tag))
		}
		if tag == "p" || tag == "div" {
			io.WriteString(w, "\n")
		}
	}
}

func filter(w io.Writer, resp *http.Response, req *http.Request) {
	if resp.StatusCode != 200 {
		resp.Write(w)
		return
	}
	if !strings.HasPrefix(resp.Header.Get("Content-Type"), "text/html") {
		resp.Write(w)
		return
	}
	log.Printf("considering filtering %s %s", req.Host, req.URL.Path)

	// this function is starting to get weird. wasteful?
	var origbuf bytes.Buffer

	io.Copy(&origbuf, resp.Body)
	resp.Body.Close()
	origbytes := origbuf.Bytes()
	bodybytes := origbytes

	gzipit := false
	if strings.EqualFold(resp.Header.Get("Content-Encoding"), "gzip") {
		var tmp bytes.Buffer
		gz, _ := gzip.NewReader(bytes.NewReader(bodybytes))
		gzipit = true
		io.Copy(&tmp, gz)
		bodybytes = tmp.Bytes()
	}

	// workaround because html doesn't parse noscript properly
	// https://github.com/golang/go/issues/16318
	{
		re := regexp.MustCompile("(?i)</?noscript[^>]*>")
		bodybytes = re.ReplaceAllLiteral(bodybytes, []byte{})
	}
	reader := bytes.NewReader(bodybytes)
	root, err := html.Parse(reader)
	if err != nil {
		log.Printf("error parsing html from %s: %s", req.URL.Path, err)
		rereader := bytes.NewReader(origbytes)
		closer := ioutil.NopCloser(rereader)
		resp.Body = closer
		resp.Write(w)
		return
	}

	L := getinterpreter()
	defer putinterpreter(L)
	var outw io.Writer
	var newbody bytes.Buffer
	if gzipit {
		outw = gzip.NewWriter(&newbody)
	} else {
		outw = &newbody
	}
	filtresult := L.Call("filter", outw, req, req.Host, req.URL.Path, root)
	if filtresult == "" {
		log.Printf("not rewriting %s", req.URL.String())
		rereader := bytes.NewReader(origbytes)
		closer := ioutil.NopCloser(rereader)
		resp.Body = closer
		resp.Write(w)
		return
	}
	if gzipit {
		zw := outw.(io.Closer)
		zw.Close()
	}

	resp.Header.Del("Content-Length")
	resp.ContentLength = -1
	closer := ioutil.NopCloser(&newbody)
	resp.Body = closer
	resp.Write(w)
}

func prefilter(req *http.Request) {
	L := getinterpreter()
	defer putinterpreter(L)
	L.Call("prefilter", req)
}

func shouldintercept(host string) bool {
	L := getinterpreter()
	defer putinterpreter(L)
	filtresult := L.Call("shouldfilter", host)
	if filtresult != "" {
		return true
	}
	return false
}
