/*
 * Copyright (c) 2017 Ted Unangst <tedu@tedunangst.com>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
package main

import (
	"bufio"
	"crypto/rand"
	"crypto/rsa"
	"crypto/tls"
	"crypto/x509"
	"crypto/x509/pkix"
	"io"
	"log"
	"math/big"
	"net"
	"net/http"
	"sync"
	"time"

	"golang.org/x/net/proxy"
)

func copyuntildone(w net.Conn, r net.Conn, pipe chan net.Conn) {
	io.Copy(w, r)
	pipe <- r
}

func closewhendone(pipe chan net.Conn) {
	w := <-pipe
	r := <-pipe
	w.Close()
	r.Close()
	close(pipe)
}

func noRedirects(req *http.Request, via []*http.Request) error {
	return http.ErrUseLastResponse
}

// get a dialer to connect out
func getDialer() proxy.Dialer {
	usetor := false

	if usetor {
		dialer, _ := proxy.SOCKS5("tcp", "127.0.0.1:9050", nil, nil)
		return dialer
	}
	return proxy.Direct
}

// this should not be different from the TLS intercept proxy
func proxyreq(w http.ResponseWriter, r *http.Request) {
	log.Printf("got a request %s\n", r.URL.String())
	if r.URL.Port() == "" {
		r.URL.Host = r.URL.Host + ":" + "80"
	}
	dialer := getDialer()
	outconn, err := dialer.Dial("tcp", r.URL.Host)
	if err != nil {
		log.Printf("err proxying %s", err)
		http.Error(w, "no proxying", http.StatusInternalServerError)
		return
	}
	r.Write(outconn)

	hj, ok := w.(http.Hijacker)
	if !ok {
		http.Error(w, "no hijacking", http.StatusInternalServerError)
		return
	}
	inconn, _, _ := hj.Hijack()
	defer inconn.Close()

	outbuf := bufio.NewReader(outconn)
	resp, err := http.ReadResponse(outbuf, r)
	if err != nil {
		log.Printf("failed to read response %s", err)
		http.Error(w, "no inner response", http.StatusInternalServerError)
		return
	}

	if !shouldintercept(r.URL.Hostname()) {
		resp.Write(inconn)
		return
	}
	filter(inconn, resp, r)
}

var certcache = make(map[string]*tls.Config)
var certexpiries = make(map[string]time.Time)
var certcachemtx sync.Mutex

func certexpiry() {
	c := time.Tick(5 * time.Minute)
	for now := range c {
		deadline := now.Add(10 * time.Minute)
		certcachemtx.Lock()
		for h, exp := range certexpiries {
			if exp.Before(deadline) {
				log.Printf("cert for %s is too old", h)
				delete(certcache, h)
				delete(certexpiries, h)
			}
		}
		certcachemtx.Unlock()
	}
}

func getTLSConfig(host string) *tls.Config {

	certcachemtx.Lock()
	tlsconf := certcache[host]
	certcachemtx.Unlock()
	if tlsconf != nil {
		return tlsconf
	}

	cacert, err := tls.LoadX509KeyPair("intercept.crt", "intercept.key")
	if err != nil {
		log.Fatalf("can't load intercept ca")
	}

	ca509, err := x509.ParseCertificate(cacert.Certificate[0])
	if err != nil {
		log.Fatalf("Failed to parse certificate: %s", err)
	}

	rsaBits := 1024
	priv, err := rsa.GenerateKey(rand.Reader, rsaBits)
	if err != nil {
		log.Fatalf("Failed to create key: %s", err)
	}

	notBefore := time.Now()
	notAfter := notBefore.Add(time.Hour)

	serialNumberLimit := new(big.Int).Lsh(big.NewInt(1), 128)
	serialNumber, err := rand.Int(rand.Reader, serialNumberLimit)
	if err != nil {
		log.Fatalf("failed to generate serial number: %s", err)
	}

	template := x509.Certificate{
		SerialNumber: serialNumber,
		Issuer:       ca509.Subject,
		Subject: pkix.Name{
			Organization: []string{"intercepted"},
		},
		NotBefore: notBefore,
		NotAfter:  notAfter,

		KeyUsage:              x509.KeyUsageKeyEncipherment | x509.KeyUsageDigitalSignature,
		ExtKeyUsage:           []x509.ExtKeyUsage{x509.ExtKeyUsageServerAuth},
		BasicConstraintsValid: true,
	}

	template.DNSNames = append(template.DNSNames, host)

	derBytes, err := x509.CreateCertificate(rand.Reader, &template, ca509, &priv.PublicKey, cacert.PrivateKey)
	if err != nil {
		log.Fatalf("Failed to create certificate: %s", err)
	}

	cert := tls.Certificate{
		Certificate: [][]byte{derBytes, cacert.Certificate[0]},
		PrivateKey:  priv,
	}

	tlsconf = &tls.Config{
		Certificates: []tls.Certificate{cert},
	}
	certcachemtx.Lock()
	certcache[host] = tlsconf
	certexpiries[host] = notAfter
	certcachemtx.Unlock()
	return tlsconf

}

type Proxer struct {
}

func (pxr *Proxer) ServeHTTP(w http.ResponseWriter, r *http.Request) {
	if r.Method == "CONNECT" {
		log.Printf("got a connect to %s", r.URL.Host)
		hj, ok := w.(http.Hijacker)
		if !ok {
			http.Error(w, "no hijacking", http.StatusInternalServerError)
			return
		}
		inconn, _, err := hj.Hijack()
		inconn.Write([]byte("HTTP/1.0 200 OK\r\n\r\n"))
		tlsconf := getTLSConfig(r.URL.Hostname())

		dialer := getDialer()
		if !shouldintercept(r.URL.Hostname()) {
			outconn, err := dialer.Dial("tcp", r.Host)
			if err != nil {
				log.Printf("failed to read request %s", err)
				http.Error(w, "no inner request", http.StatusInternalServerError)
				return
			}
			p := make(chan net.Conn)
			go copyuntildone(inconn, outconn, p)
			go copyuntildone(outconn, inconn, p)
			go closewhendone(p)
			return
		}
		inconn = tls.Server(inconn, tlsconf)
		defer inconn.Close()

		inbuf := bufio.NewReader(inconn)
		inreq, err := http.ReadRequest(inbuf)
		if err != nil {
			log.Printf("failed to read request %s", err)
			http.Error(w, "no inner request", http.StatusInternalServerError)
			return
		}
		inreq.Host = ""
		inreq.URL.Scheme = "https"
		inreq.URL.Host = r.Host
		inreq.Close = true
		inreq.Header["Connection"] = []string{"close"}
		log.Printf("inreq url %s\n", inreq.URL.String())
		prefilter(inreq)

		log.Printf("dialing %s\n", inreq.URL.Host)
		rawconn, err := dialer.Dial("tcp", inreq.URL.Host)
		if err != nil {
			log.Printf("failed to connect %s", err)
			http.Error(w, "no connecting", http.StatusInternalServerError)
			return
		}
		outconn := tls.Client(rawconn, &tls.Config{ServerName: inreq.URL.Hostname()})
		inreq.URL.Host = inreq.URL.Hostname()
		defer outconn.Close()
		log.Printf("connection success")
		inreq.Write(outconn)
		outbuf := bufio.NewReader(outconn)
		resp, err := http.ReadResponse(outbuf, inreq)
		if err != nil {
			log.Printf("failed to read response %s", err)
			http.Error(w, "no inner response", http.StatusInternalServerError)
			return
		}
		resp.Close = true
		resp.Header["Connection"] = []string{"close"}
		filter(inconn, resp, inreq)
		return
	}
	proxyreq(w, r)
}

func main() {
	go certexpiry()
	var pxr Proxer
	http.ListenAndServe("127.0.0.1:8090", &pxr)
}
